/usr/bin/time -v python bench/ndarray/jit-reduc-sizes.py numpy
/usr/bin/time -v python bench/ndarray/jit-reduc-sizes.py numpy_jit
/usr/bin/time -v python bench/ndarray/jit-reduc-sizes.py 0
/usr/bin/time -v python bench/ndarray/jit-reduc-sizes.py 1 LZ4
/usr/bin/time -v python bench/ndarray/jit-reduc-sizes.py 1 ZSTD
